package org.openjdk.gcbench.fragger;

import org.openjdk.gcbench.util.AllocProfileSupport;
import org.openjdk.gcbench.util.ratelimit.MultiTokenBucket;
import org.openjdk.jmh.annotations.*;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(value = 1)
@BenchmarkMode(Mode.Throughput)
@OutputTimeUnit(TimeUnit.SECONDS)
@Threads(Threads.MAX)
@State(Scope.Benchmark)
public class LinkedListFragger {

    public static final int PAYLOAD_SIZE = 16384;

    @Param({"1000"})
    long ldsMB;

    @Param({"10"})
    int rate;

    List<Object> objects;

    MultiTokenBucket bucket;

    int count;

    LinkedList<Object> o0;
    LinkedList<Object> o16;

    private int getSizePerCount() {
        if (AllocProfileSupport.isAvailable()) {
            long a1 = AllocProfileSupport.getAllocatedBytes();
            o0 = new LinkedList<>();
            long a2 = AllocProfileSupport.getAllocatedBytes();
            o16 = new LinkedList<>();
            for (int c = 0; c < 16; c++) {
                o16.add(new byte[PAYLOAD_SIZE]);
            }
            long a3 = AllocProfileSupport.getAllocatedBytes();

            long sizeO0 = a2 - a1;
            long sizeO16 = a3 - a2;

            return (int) ((sizeO16 - sizeO0) / 16);
        } else {
            // best guess
            return 24 + // Node
                    16 + PAYLOAD_SIZE; // payload
        }
    }

    @Setup
    public void setup() {
        bucket = new MultiTokenBucket(rate);
        count = (int)Math.max(1, ldsMB * 1024 * 1024 / getSizePerCount());
        objects = new LinkedList<>();
        for (int c = 0; c < count; c++) {
            objects.add(new byte[PAYLOAD_SIZE]);
        }
    }

    @Benchmark
    public void test() {
        bucket.limit();
        objects.set(ThreadLocalRandom.current().nextInt(count), new byte[PAYLOAD_SIZE]);
    }

}
