package org.openjdk.gcbench.roots;

import org.openjdk.jmh.annotations.*;
import org.openjdk.jmh.infra.Blackhole;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(value = 1, jvmArgsAppend = "-Xss32m")
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.MICROSECONDS)
@Threads(Threads.MAX)
@State(Scope.Thread)
public class Locals {

    List<Object> list;

    @Param({"100000"})
    private int size;

    @Setup
    public void setup() {
        size = Math.max(100, size);
        list = new ArrayList<>();
        for (int c = 0; c < size; c++) {
            list.add(new Object());
        }
    }

    @Benchmark
    public void test(Blackhole bh) throws InterruptedException {
        recursiveLocal(bh, list, 10);
    }

    @CompilerControl(CompilerControl.Mode.DONT_INLINE)
    private void recursiveLocal(Blackhole bh, List<Object> list, int i) {
        bh.consume(new int[64]); // Allocation pressure to trigger GCs.

        Object o00, o01, o02, o03, o04, o05, o06, o07, o08, o09;
        o00 = o01 = o02 = o03 = o04 = o05 = o06 = o07 = o08 = o09 = null;
        if (i < list.size()) {
            o00 = list.get(i - 0);
            o01 = list.get(i - 1);
            o02 = list.get(i - 2);
            o03 = list.get(i - 3);
            o04 = list.get(i - 4);
            o05 = list.get(i - 5);
            o06 = list.get(i - 6);
            o07 = list.get(i - 7);
            o08 = list.get(i - 8);
            o09 = list.get(i - 9);
            recursiveLocal(bh, list, i + 1);
        }
        bh.consume(o00);
        bh.consume(o01);
        bh.consume(o02);
        bh.consume(o03);
        bh.consume(o04);
        bh.consume(o05);
        bh.consume(o06);
        bh.consume(o07);
        bh.consume(o08);
        bh.consume(o09);
    }

}
